from __future__ import unicode_literals

__author__ = "Shyue Ping Ong, William Davidson Richards, Stephen Dacek, " \
             "Xiaohui Qu"
__date__ = "Sep 22 2014"
__version__ = "0.8.0"

from .custodian import Custodian
